@@ -1,28 +1,3 @@
-/*
-  Racket
-  Copyright (c) 2004-2016 PLT Design Inc.
-  Copyright (c) 1995-2001 Matthew Flatt
-
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Library General Public
-    License as published by the Free Software Foundation; either
-    version 2 of the License, or (at your option) any later version.
-
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Library General Public License for more details.
-
-    You should have received a copy of the GNU Library General Public
-    License along with this library; if not, write to the Free
-    Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-    Boston, MA 02110-1301 USA.
-
-  libscheme
-  Copyright (c) 1994 Brent Benson
-  All rights reserved.
-*/
-
 #include "schpriv.h"
 #include "schrunst.h"
 
@@ -33,10 +8,14 @@ SHARED_OK Scheme_Hash_Tree *empty_hash_tree;
 
 SHARED_OK static int validate_compile_result = 0;
 SHARED_OK static int recompile_every_compile = 0;
+SHARED_OK static int show_linklets = 0;
 
 static Scheme_Object *serializable_symbol;
 static Scheme_Object *unsafe_symbol;
 static Scheme_Object *static_symbol;
+static Scheme_Object *use_prompt_symbol;
+static Scheme_Object *uninterned_literal_symbol;
+static Scheme_Object *quick_symbol;
 static Scheme_Object *constant_symbol;
 static Scheme_Object *consistent_symbol;
 static Scheme_Object *noncm_symbol;
@@ -45,6 +24,9 @@ static Scheme_Object *omitable_symbol;
 static Scheme_Object *folding_symbol;
 
 THREAD_LOCAL_DECL(Scheme_Hash_Table *local_primitive_tables);
+THREAD_LOCAL_DECL(extern intptr_t scheme_code_page_total);
+THREAD_LOCAL_DECL(extern intptr_t scheme_code_total);
+THREAD_LOCAL_DECL(extern intptr_t scheme_code_count);
 
 static Scheme_Object *primitive_table(int argc, Scheme_Object **argv);
 static Scheme_Object *primitive_to_position(int argc, Scheme_Object **argv);
@@ -55,11 +37,14 @@ static Scheme_Object *linklet_p(int argc, Scheme_Object **argv);
 static Scheme_Object *compile_linklet(int argc, Scheme_Object **argv);
 static Scheme_Object *recompile_linklet(int argc, Scheme_Object **argv);
 static Scheme_Object *eval_linklet(int argc, Scheme_Object **argv);
-static Scheme_Object *read_compiled_linklet(int argc, Scheme_Object **argv);
 static Scheme_Object *instantiate_linklet(int argc, Scheme_Object **argv);
 static Scheme_Object *linklet_import_variables(int argc, Scheme_Object **argv);
 static Scheme_Object *linklet_export_variables(int argc, Scheme_Object **argv);
 
+static Scheme_Object *linklet_vm_bytes(int argc, Scheme_Object **argv);
+static Scheme_Object *write_linklet_bundle_hash(int argc, Scheme_Object **argv);
+static Scheme_Object *read_linklet_bundle_hash(int argc, Scheme_Object **argv);
+
 static Scheme_Object *instance_p(int argc, Scheme_Object **argv);
 static Scheme_Object *make_instance(int argc, Scheme_Object **argv);
 static Scheme_Object *instance_name(int argc, Scheme_Object **argv);
@@ -68,14 +53,7 @@ static Scheme_Object *instance_variable_names(int argc, Scheme_Object **argv);
 static Scheme_Object *instance_variable_value(int argc, Scheme_Object **argv);
 static Scheme_Object *instance_set_variable_value(int argc, Scheme_Object **argv);
 static Scheme_Object *instance_unset_variable(int argc, Scheme_Object **argv);
-
-static Scheme_Object *linklet_directory_p(int argc, Scheme_Object **argv);
-static Scheme_Object *linklet_directory_to_hash(int argc, Scheme_Object **argv);
-static Scheme_Object *hash_to_linklet_directory(int argc, Scheme_Object **argv);
-
-static Scheme_Object *linklet_bundle_p(int argc, Scheme_Object **argv);
-static Scheme_Object *linklet_bundle_to_hash(int argc, Scheme_Object **argv);
-static Scheme_Object *hash_to_linklet_bundle(int argc, Scheme_Object **argv);
+static Scheme_Object *instance_describe_variable(int argc, Scheme_Object **argv);
 
 static Scheme_Object *variable_p(int argc, Scheme_Object **argv);
 static Scheme_Object *variable_instance(int argc, Scheme_Object **argv);
@@ -86,7 +64,7 @@ static Scheme_Linklet *compile_and_or_optimize_linklet(Scheme_Object *form, Sche
                                                        Scheme_Object *name,
                                                        Scheme_Object **_import_keys,
                                                        Scheme_Object *get_import,
-                                                       int unsafe_mode, int static_mode);
+                                                       int unsafe_mode, int static_mode, int serializable);
 
 static Scheme_Object *_instantiate_linklet_multi(Scheme_Linklet *linklet, Scheme_Instance *instance,
                                                  int num_instances, Scheme_Instance **instances,
@@ -123,9 +101,15 @@ void scheme_init_linklet(Scheme_Startup_Env *env)
   REGISTER_SO(serializable_symbol);
   REGISTER_SO(unsafe_symbol);
   REGISTER_SO(static_symbol);
+  REGISTER_SO(use_prompt_symbol);
+  REGISTER_SO(uninterned_literal_symbol);
+  REGISTER_SO(quick_symbol);
   serializable_symbol = scheme_intern_symbol("serializable");
   unsafe_symbol = scheme_intern_symbol("unsafe");
   static_symbol = scheme_intern_symbol("static");
+  use_prompt_symbol = scheme_intern_symbol("use-prompt");
+  uninterned_literal_symbol = scheme_intern_symbol("uninterned-literal");
+  quick_symbol = scheme_intern_symbol("quick");
 
   REGISTER_SO(constant_symbol);
   REGISTER_SO(consistent_symbol);
@@ -149,13 +133,16 @@ void scheme_init_linklet(Scheme_Startup_Env *env)
 
   ADD_FOLDING_PRIM("linklet?", linklet_p, 1, 1, 1, env);
   ADD_PRIM_W_ARITY2("compile-linklet", compile_linklet, 1, 5, 2, 2, env);
-  ADD_PRIM_W_ARITY2("recompile-linklet", recompile_linklet, 1, 4, 2, 2, env);
+  ADD_PRIM_W_ARITY2("recompile-linklet", recompile_linklet, 1, 5, 2, 2, env);
   ADD_IMMED_PRIM("eval-linklet", eval_linklet, 1, 1, env);
-  ADD_PRIM_W_ARITY("read-compiled-linklet", read_compiled_linklet, 1, 1, env);
   ADD_PRIM_W_ARITY2("instantiate-linklet", instantiate_linklet, 2, 4, 0, -1, env);
   ADD_PRIM_W_ARITY("linklet-import-variables", linklet_import_variables, 1, 1, env);
   ADD_PRIM_W_ARITY("linklet-export-variables", linklet_export_variables, 1, 1, env);
 
+  ADD_PRIM_W_ARITY("linklet-virtual-machine-bytes", linklet_vm_bytes, 0, 0, env);
+  ADD_PRIM_W_ARITY("write-linklet-bundle-hash", write_linklet_bundle_hash, 2, 2, env);
+  ADD_PRIM_W_ARITY("read-linklet-bundle-hash", read_linklet_bundle_hash, 1, 1, env);
+
   ADD_FOLDING_PRIM("instance?", instance_p, 1, 1, 1, env);
   ADD_PRIM_W_ARITY("make-instance", make_instance, 1, -1, env);
   ADD_PRIM_W_ARITY("instance-name", instance_name, 1, 1, env);
@@ -164,14 +151,7 @@ void scheme_init_linklet(Scheme_Startup_Env *env)
   ADD_PRIM_W_ARITY2("instance-variable-value", instance_variable_value, 2, 3, 0, -1, env);
   ADD_PRIM_W_ARITY("instance-set-variable-value!", instance_set_variable_value, 3, 4, env);
   ADD_PRIM_W_ARITY("instance-unset-variable!", instance_unset_variable, 2, 2, env);
-
-  ADD_FOLDING_PRIM("linklet-directory?", linklet_directory_p, 1, 1, 1, env);
-  ADD_PRIM_W_ARITY("hash->linklet-directory", hash_to_linklet_directory, 1, 1, env);
-  ADD_PRIM_W_ARITY("linklet-directory->hash", linklet_directory_to_hash, 1, 1, env);
-
-  ADD_FOLDING_PRIM("linklet-bundle?", linklet_bundle_p, 1, 1, 1, env);
-  ADD_PRIM_W_ARITY("hash->linklet-bundle", hash_to_linklet_bundle, 1, 1, env);
-  ADD_PRIM_W_ARITY("linklet-bundle->hash", linklet_bundle_to_hash, 1, 1, env);
+  ADD_PRIM_W_ARITY("instance-describe-variable!", instance_describe_variable, 3, 3, env);
 
   ADD_FOLDING_PRIM_UNARY_INLINED("variable-reference?", variable_p, 1, 1, 1, env);
   ADD_IMMED_PRIM("variable-reference->instance", variable_instance, 1, 2, env);
@@ -213,6 +193,9 @@ void scheme_init_linklet(Scheme_Startup_Env *env)
         recompile_every_compile = 32;
     }
   }
+
+  if (scheme_getenv("PLT_LINKLET_SHOW"))
+    show_linklets = 1;
 }
 
 void scheme_init_unsafe_linklet(Scheme_Startup_Env *env)
@@ -315,9 +298,10 @@ static Scheme_Object *primitive_in_category_p(int argc, Scheme_Object **argv)
     else if (SAME_OBJ(cat, folding_symbol))
       r = (opt >= SCHEME_PRIM_OPT_FOLDING);
     else if (SAME_OBJ(cat, omitable_symbol))
-      r = (SCHEME_PRIM_PROC_OPT_FLAGS(v) & (SCHEME_PRIM_IS_OMITABLE_ANY
-                                            | SCHEME_PRIM_IS_UNSAFE_FUNCTIONAL
-                                            | SCHEME_PRIM_IS_UNSAFE_OMITABLE));
+      r = (SCHEME_PRIM_PROC_OPT_FLAGS(v) & (SCHEME_PRIM_IS_OMITABLE
+                                            | SCHEME_PRIM_IS_OMITABLE_ALLOCATION
+                                            | SCHEME_PRIM_IS_UNSAFE_OMITABLE
+                                            | SCHEME_PRIM_IS_UNSAFE_FUNCTIONAL));
     else
       r = 0;
   } else
@@ -361,7 +345,7 @@ void extract_import_info(const char *who, int argc, Scheme_Object **argv,
       if (!*_import_keys) {
         scheme_contract_error(who,
                               "no vector supplied for import keys, but import-getting function provided;\n"
-                              " the function argument must be `#f' when the vector argument is `#f'",
+                              " the function argument must be `#f` when the vector argument is `#f`",
                               "import-getting function", 1, argv[3],
                               NULL);
       }
@@ -372,12 +356,70 @@ void extract_import_info(const char *who, int argc, Scheme_Object **argv,
     *_get_import = NULL;
 }
 
+static void parse_compile_options(const char *who, int arg_pos,
+                                  int argc, Scheme_Object **argv,
+                                  int *_unsafe, int *_static_mode,
+                                  int *_serializable)
+{
+  Scheme_Object *redundant = NULL, *flag, *flags = argv[arg_pos];
+  int serializable = 0;
+  int unsafe = *_unsafe;
+  int static_mode = *_static_mode;
+  int use_prompt_mode = 0;
+  int uninterned_literal_mode = 0;
+  int quick_mode = 0;
+  
+  while (SCHEME_PAIRP(flags)) {
+    flag = SCHEME_CAR(flags);
+    if (SAME_OBJ(flag, serializable_symbol)) {
+      if (serializable && !redundant)
+        redundant = flag;
+      serializable = 1;
+    } else if (SAME_OBJ(flag, unsafe_symbol)) {
+      if (unsafe && !redundant)
+        redundant = flag;
+      unsafe = 1;
+    } else if (SAME_OBJ(flag, static_symbol)) {
+      if (static_mode && !redundant)
+        redundant = flag;
+      static_mode = 1;
+    } else if (SAME_OBJ(flag, use_prompt_symbol)) {
+      if (use_prompt_mode && !redundant)
+        redundant = flag;
+      use_prompt_mode = 1;
+    } else if (SAME_OBJ(flag, uninterned_literal_symbol)) {
+      if (uninterned_literal_mode && !redundant)
+        redundant = flag;
+      uninterned_literal_mode = 1;
+    } else if (SAME_OBJ(flag, quick_symbol)) {
+      if (quick_mode && !redundant)
+        redundant = flag;
+      quick_mode = 1;
+    } else
+      break;
+    flags = SCHEME_CDR(flags);
+  }
+
+  if (!SCHEME_NULLP(flags))
+    scheme_wrong_contract("compile-linklet",
+                          "(listof/c 'serializable 'unsafe 'static 'use-prompt 'uninterned-literal 'quick)",
+                          arg_pos, argc, argv);
+
+  if (redundant)
+    scheme_contract_error("compile-linklet", "redundant option",
+                          "redundant option", 1, redundant,
+                          "supplied options", 1, argv[arg_pos],
+                          NULL);
+
+  *_unsafe = unsafe;
+  *_static_mode = static_mode;
+  *_serializable = serializable;
+}
+
 static Scheme_Object *compile_linklet(int argc, Scheme_Object **argv)
 {
   Scheme_Object *name, *e, *import_keys, *get_import, *a[2];
-  int unsafe = 0, static_mode = 0;
-
-  /* Last argument, `serializable?`, is ignored */
+  int unsafe = 0, static_mode = 0, serializable = 1;
 
   extract_import_info("compile-linklet", argc, argv, &import_keys, &get_import);
 
@@ -390,40 +432,18 @@ static Scheme_Object *compile_linklet(int argc, Scheme_Object **argv)
   if (!SCHEME_STXP(e))
     e = scheme_datum_to_syntax(e, scheme_false, DTS_CAN_GRAPH);
 
-  if (argc > 4) {
-    Scheme_Object *flags, *redundant = NULL, *flag;
-    int serializable = 0;
-      
-    flags = argv[4];
-    while (SCHEME_PAIRP(flags)) {
-      flag = SCHEME_CAR(flags);
-      if (SAME_OBJ(flag, serializable_symbol)) {
-        if (serializable && !redundant)
-          redundant = flag;
-        serializable = 1;
-      } else if (SAME_OBJ(flag, unsafe_symbol)) {
-        if (unsafe && !redundant)
-          redundant = flag;
-        unsafe = 1;
-      } else if (SAME_OBJ(flag, static_symbol)) {
-        if (static_mode && !redundant)
-          redundant = flag;
-        static_mode = 1;
-      } else
-        break;
-      flags = SCHEME_CDR(flags);
-    }
-    if (!SCHEME_NULLP(flags))
-      scheme_wrong_contract("compile-linklet", "(listof/c 'serializable 'unsafe)", 4, argc, argv);
-    if (redundant)
-      scheme_contract_error("compile-linklet", "redundant option",
-                            "redundant option", 1, redundant,
-                            "supplied options", 1, argv[4],
-                            NULL);
+  if (show_linklets) {
+    char *s;
+    intptr_t s_len;
+    s = scheme_write_to_string(scheme_syntax_to_datum(e), &s_len);
+    printf("%s\n", s);
   }
 
+  if (argc > 4)
+    parse_compile_options("compile-linklet", 4, argc, argv, &unsafe, &static_mode, &serializable);
+
   e = (Scheme_Object *)compile_and_or_optimize_linklet(e, NULL, name, &import_keys, get_import,
-                                                       unsafe, static_mode);
+                                                       unsafe, static_mode, serializable);
 
   if (import_keys) {
     a[0] = e;
@@ -437,7 +457,8 @@ static Scheme_Object *recompile_linklet(int argc, Scheme_Object **argv)
 {
   Scheme_Object *name, *import_keys, *get_import, *a[2];
   Scheme_Linklet *linklet;
-    
+  int unsafe = 0, static_mode = 0, serializable = 1;
+
   if (!SAME_TYPE(SCHEME_TYPE(argv[0]), scheme_linklet_type))
     scheme_wrong_contract("recompile-linklet", "linklet?", 0, argc, argv);
 
@@ -460,8 +481,12 @@ static Scheme_Object *recompile_linklet(int argc, Scheme_Object **argv)
                           "given keys", 1, scheme_make_integer(SCHEME_VEC_SIZE(import_keys)),
                           NULL);
   }
+
+  if (argc > 4)
+    parse_compile_options("recompile-linklet", 4, argc, argv, &unsafe, &static_mode, &serializable);
   
-  linklet = compile_and_or_optimize_linklet(NULL, linklet, name, &import_keys, get_import, 0, 0);
+  linklet = compile_and_or_optimize_linklet(NULL, linklet, name, &import_keys, get_import,
+                                            unsafe, static_mode, serializable);
 
   if (import_keys) {
     a[0] = (Scheme_Object *)linklet;
@@ -497,15 +522,72 @@ static Scheme_Object *eval_linklet(int argc, Scheme_Object **argv)
     }
   }
 
+#ifdef MZ_USE_JIT
+  if (linklet->native_lambdas) {
+    Scheme_Object *l;
+    l = linklet->native_lambdas;
+    linklet->native_lambdas = NULL;
+    while (SCHEME_PAIRP(l)) {
+      scheme_force_jit_generate((Scheme_Native_Lambda *)SCHEME_CAR(l));
+      l = SCHEME_CDR(l);
+    }
+  }
+#endif
+
   return (Scheme_Object *)linklet;
 }
 
-static Scheme_Object *read_compiled_linklet(int argc, Scheme_Object **argv)
+static Scheme_Object *linklet_vm_bytes(int argc, Scheme_Object **argv)
+{
+  return scheme_make_byte_string("racket");
+}
+
+static Scheme_Object *read_linklet_bundle_hash(int argc, Scheme_Object **argv)
 {
   if (!SCHEME_INPUT_PORTP(argv[0]))
-    scheme_wrong_contract("read-compiled-linklet", "input-port?", 0, argc, argv);
+    scheme_wrong_contract("read-linklet-bundle-hash", "input-port?", 0, argc, argv);
 
-  return scheme_read_compiled(argv[0]);
+  return scheme_read_linklet_bundle_hash(argv[0]);
+}
+
+static Scheme_Object *write_linklet_bundle_hash(int argc, Scheme_Object **argv)
+{
+  mzlonglong pos;
+  Scheme_Object *k, *v;
+  Scheme_Hash_Tree *hash;
+  
+  if (!SCHEME_HASHTRP(argv[0])
+      || !SAME_TYPE(scheme_eq_hash_tree_type, SCHEME_HASHTR_TYPE(argv[0])))
+    scheme_wrong_contract("write-linklet-bundle-hash",
+                          "(and/c hash? hash-eq? immutable? (not/c impersonator?))",
+                          0, argc, argv);
+
+  if (!SCHEME_OUTPUT_PORTP(argv[1]))
+    scheme_wrong_contract("write-linklet-bundle-hash", "output-port?", 0, argc, argv);
+
+  hash = (Scheme_Hash_Tree *)argv[0];
+
+  /* mapping: keys must be symbols and fixnums */
+
+  pos = scheme_hash_tree_next(hash, -1);
+  while (pos != -1) {
+    scheme_hash_tree_index(hash, pos, &k, &v);
+    if (!SCHEME_SYMBOLP(k) && !SCHEME_INTP(k)) {
+      scheme_contract_error("write-linklet-bundle-hash",
+                            "key in given hash is not a symbol or fixnum",
+                            "key", 1, k,
+                            NULL);
+    }
+    pos = scheme_hash_tree_next(hash, pos);
+  }
+
+  v = scheme_alloc_small_object();
+  v->type = scheme_linklet_bundle_type;
+  SCHEME_PTR_VAL(v) = argv[0];
+
+  scheme_write(v, argv[1]);
+
+  return scheme_void;
 }
 
 static Scheme_Object *instantiate_linklet(int argc, Scheme_Object **argv)
@@ -519,7 +601,7 @@ static Scheme_Object *instantiate_linklet(int argc, Scheme_Object **argv)
     scheme_wrong_contract("instantiate-linklet", "linklet?", 0, argc, argv);
 
   l = argv[1];
-  while (!SCHEME_NULLP(l)) {
+  while (SCHEME_PAIRP(l)) {
     if (!SAME_TYPE(SCHEME_TYPE(SCHEME_CAR(l)), scheme_instance_type))
       break;
     l = SCHEME_CDR(l);
@@ -780,114 +862,14 @@ static Scheme_Object *instance_unset_variable(int argc, Scheme_Object **argv)
   return scheme_void;
 }
 
-static Scheme_Object *linklet_directory_p(int argc, Scheme_Object **argv)
+static Scheme_Object *instance_describe_variable(int argc, Scheme_Object **argv)
 {
-  return (SAME_TYPE(SCHEME_TYPE(argv[0]), scheme_linklet_directory_type)
-          ? scheme_true
-          : scheme_false);
-}
+  if (!SAME_TYPE(SCHEME_TYPE(argv[0]), scheme_instance_type))
+    scheme_wrong_contract("instance-describe-variable!", "instance?", 0, argc, argv);
+  if (!SCHEME_SYMBOLP(argv[1]))
+    scheme_wrong_contract("instance-describe-variable!", "symbol?", 1, argc, argv);
 
-static Scheme_Object *linklet_directory_to_hash(int argc, Scheme_Object **argv)
-{
-  if (!SAME_TYPE(SCHEME_TYPE(argv[0]), scheme_linklet_directory_type))
-    scheme_wrong_contract("linklet-directory->hash", "linklet-directory?", 0, argc, argv);
-
-  return SCHEME_PTR_VAL(argv[0]);
-}
-
-static Scheme_Object *hash_to_linklet_directory(int argc, Scheme_Object **argv)
-{
-  mzlonglong pos;
-  Scheme_Object *k, *v;
-  Scheme_Hash_Tree *hash;
-  
-  if (!SCHEME_HASHTRP(argv[0])
-      || !SAME_TYPE(scheme_eq_hash_tree_type, SCHEME_HASHTR_TYPE(argv[0])))
-    scheme_wrong_contract("hash->linklet-directory",
-                          "(and/c hash? hash-eq? immutable? (not/c impersonator?))",
-                          0, argc, argv);
-  hash = (Scheme_Hash_Tree *)argv[0];
-
-  /* mapping: #f -> bundle, sym -> linklet directory */
-
-  pos = scheme_hash_tree_next(hash, -1);
-  while (pos != -1) {
-    scheme_hash_tree_index(hash, pos, &k, &v);
-    if (SCHEME_FALSEP(k)) {
-      if (!SAME_TYPE(SCHEME_TYPE(v), scheme_linklet_bundle_type))
-        scheme_contract_error("hash->linklet-directory",
-                              "value for #f key is not a linklet bundle",
-                              "value", 1, v,
-                              NULL);
-    } else if (SCHEME_SYMBOLP(k)) {
-      if (!SAME_TYPE(SCHEME_TYPE(v), scheme_linklet_directory_type))
-        scheme_contract_error("hash->linklet-directory",
-                              "value for symbol key is not a linklet directory",
-                              "key", 1, k,
-                              "value", 1, v,
-                              NULL);
-    } else {
-      scheme_contract_error("hash->linklet-directory",
-                            "key in given hash is not #f or a symbol",
-                            "key", 1, k,
-                            NULL);
-    }
-    pos = scheme_hash_tree_next(hash, pos);
-  }
-
-  v = scheme_alloc_small_object();
-  v->type = scheme_linklet_directory_type;
-  SCHEME_PTR_VAL(v) = argv[0];
-  return v;
-}
-
-static Scheme_Object *linklet_bundle_p(int argc, Scheme_Object **argv)
-{
-  return (SAME_TYPE(SCHEME_TYPE(argv[0]), scheme_linklet_bundle_type)
-          ? scheme_true
-          : scheme_false);
-}
-
-static Scheme_Object *linklet_bundle_to_hash(int argc, Scheme_Object **argv)
-{
-  if (!SAME_TYPE(SCHEME_TYPE(argv[0]), scheme_linklet_bundle_type))
-    scheme_wrong_contract("linklet-bundle->hash", "linklet-bundle?", 0, argc, argv);
-
-  return SCHEME_PTR_VAL(argv[0]);
-}
-
-static Scheme_Object *hash_to_linklet_bundle(int argc, Scheme_Object **argv)
-{
-  mzlonglong pos;
-  Scheme_Object *k, *v;
-  Scheme_Hash_Tree *hash;
-  
-  if (!SCHEME_HASHTRP(argv[0])
-      || !SAME_TYPE(scheme_eq_hash_tree_type, SCHEME_HASHTR_TYPE(argv[0])))
-    scheme_wrong_contract("hash->linklet-bundle",
-                          "(and/c hash? hash-eq? immutable? (not/c impersonator?))",
-                          0, argc, argv);
-
-  hash = (Scheme_Hash_Tree *)argv[0];
-
-  /* mapping: keys must be symbols and fixnums */
-
-  pos = scheme_hash_tree_next(hash, -1);
-  while (pos != -1) {
-    scheme_hash_tree_index(hash, pos, &k, &v);
-    if (!SCHEME_SYMBOLP(k) && !SCHEME_INTP(k)) {
-      scheme_contract_error("hash->linklet-bundle",
-                            "key in given hash is not a symbol or fixnum",
-                            "key", 1, k,
-                            NULL);
-    }
-    pos = scheme_hash_tree_next(hash, pos);
-  }
-
-  v = scheme_alloc_small_object();
-  v->type = scheme_linklet_bundle_type;
-  SCHEME_PTR_VAL(v) = argv[0];
-  return v;
+  return scheme_void;
 }
 
 static Scheme_Object *variable_p(int argc, Scheme_Object **argv)
@@ -1162,10 +1144,13 @@ static Scheme_Hash_Tree *update_source_names(Scheme_Hash_Tree *source_names,
 static Scheme_Linklet *compile_and_or_optimize_linklet(Scheme_Object *form, Scheme_Linklet *linklet,
                                                        Scheme_Object *name,
                                                        Scheme_Object **_import_keys, Scheme_Object *get_import,
-                                                       int unsafe_mode, int static_mode)
+                                                       int unsafe_mode, int static_mode, int serializable)
 {
   Scheme_Config *config;
   int enforce_const, set_undef, can_inline;
+  Scheme_Performance_State perf_state;
+
+  scheme_performance_record_start(&perf_state);
 
   config = scheme_current_config();
   enforce_const = SCHEME_TRUEP(scheme_get_param(config, MZCONFIG_COMPILE_MODULE_CONSTS));
@@ -1202,12 +1187,17 @@ static Scheme_Linklet *compile_and_or_optimize_linklet(Scheme_Object *form, Sche
   if (validate_compile_result)
     scheme_validate_linklet(NULL, linklet);
 
+  scheme_performance_record_end("compile", &perf_state);
+
+  if (serializable)
+    linklet->serializable = 1;
+
   return linklet;
 }
 
 Scheme_Linklet *scheme_compile_and_optimize_linklet(Scheme_Object *form, Scheme_Object *name)
 {
-  return compile_and_or_optimize_linklet(form, NULL, name, NULL, NULL, 0, 1);
+  return compile_and_or_optimize_linklet(form, NULL, name, NULL, NULL, 0, 1, 0);
 }
 
 /*========================================================================*/
@@ -1285,7 +1275,7 @@ Scheme_Object *scheme_linklet_run_finish(Scheme_Linklet* linklet, Scheme_Instanc
         /* Double-check that the definition-installing part of the
            continuation was not skipped. Otherwise, the compiler would
            not be able to assume that a variable reference that is
-           lexically later (incuding a reference to an imported
+           lexically later (including a reference to an imported
            variable) always references a defined variable. Putting the
            prompt around a definition's RHS might be a better
            approach, but that would change the language (so mabe next
@@ -1351,6 +1341,7 @@ static void *instantiate_linklet_k(void)
   int depth;
   Scheme_Object *b, *v;
   Scheme_Hash_Tree *source_names;
+  Scheme_Performance_State perf_state;
 
   p->ku.k.p1 = NULL;
   p->ku.k.p2 = NULL;
@@ -1367,6 +1358,8 @@ static void *instantiate_linklet_k(void)
     return (Scheme_Object *)scheme_enlarge_runstack(depth, instantiate_linklet_k);
   }
 
+  scheme_performance_record_start(&perf_state);
+
   if (!linklet->jit_ready) {
     b = scheme_get_param(scheme_current_config(), MZCONFIG_USE_JIT);
     if (SCHEME_TRUEP(b))
@@ -1403,6 +1396,40 @@ static void *instantiate_linklet_k(void)
   if (!multi)
     v = scheme_check_one_value(v);
 
+#ifdef MZ_USE_JIT
+  if (linklet->native_lambdas) {
+    int mc;
+    Scheme_Object **mv, *l;
+
+    if (SAME_OBJ(v, SCHEME_MULTIPLE_VALUES)) {
+      p = scheme_current_thread;
+      mv = p->ku.multiple.array;
+      mc = p->ku.multiple.count;
+      if (SAME_OBJ(mv, p->values_buffer))
+        p->values_buffer = NULL;
+    } else {
+      mv = NULL;
+      mc = 0;
+    }
+
+    l = linklet->native_lambdas;
+    linklet->native_lambdas = NULL;
+
+    while (SCHEME_PAIRP(l)) {
+      scheme_force_jit_generate((Scheme_Native_Lambda *)SCHEME_CAR(l));
+      l = SCHEME_CDR(l);
+    }
+
+    if (mv) {
+      p = scheme_current_thread;
+      p->ku.multiple.array = mv;
+      p->ku.multiple.count = mc;
+    }
+  }
+#endif
+
+  scheme_performance_record_end("instantiate", &perf_state);
+
   return (void *)v;
 }
 
@@ -1510,7 +1537,7 @@ static Scheme_Hash_Tree *push_prefix(Scheme_Linklet *linklet, Scheme_Instance *i
 
       if (v) {
         if (!((Scheme_Bucket *)v)->val) {
-          bad_reason = "is unintialized";
+          bad_reason = "is uninitialized";
           v = NULL;
         } else if (linklet->import_shapes) {
           Scheme_Object *shape = SCHEME_VEC_ELS(linklet->import_shapes)[pos-1];
@@ -1713,6 +1740,309 @@ int check_pruned_prefix(void *p) XFORM_SKIP_PROC
 }
 #endif
 
+/*========================================================================*/
+/*  Recorindg performance times                                           */
+/*========================================================================*/
+
+static intptr_t nested_delta, nested_gc_delta;
+static int perf_reg, perf_count;
+
+typedef struct {
+  const char *name;
+  intptr_t accum;
+  intptr_t gc_accum;
+  intptr_t count;
+} Performance_Entry;
+
+#define MAX_PERF_ENTRIES 16
+
+static Performance_Entry perf_entries[MAX_PERF_ENTRIES];
+
+#define MAX_PERF_CATS    3
+#define MAX_PERF_SUBS    3
+
+typedef struct {
+  const char *name;
+  Performance_Entry perf_entries[MAX_PERF_SUBS];
+  int perf_count;
+} Performance_Cat;
+
+typedef struct {
+  const char *entry;
+  const char *cat;
+} Performace_Recat;
+
+static Performace_Recat recats[] = { { "instantiate", "run" },
+                                     { "jit", "run" },
+                                     { "comp-ffi-call", "comp-ffi" },
+                                     { "comp-ffi-back", "comp-ffi" },
+                                     { NULL, NULL} };
+
+static char *do_tab(int len, char *tab, int max_len)
+{
+  int i;
+
+  len = max_len - len;
+  if (len < 0)
+    len = 0;
+  for (i = 0; i < len; i++) {
+    tab[i] = ' ';
+  }
+  tab[i] = 0;
+
+  return tab;
+}
+
+static int numlen(intptr_t n)
+{
+  int len = 1;
+
+  while (n >= 10) {
+    n = n / 10;
+    len++;
+  }
+
+  return len;
+}
+
+static char *tab_number(intptr_t n, char *tab, int max_len)
+{
+  return do_tab(numlen(n), tab, max_len);
+}
+
+static char *tab_string(const char *s, char *tab, int max_len)
+{
+  return do_tab(strlen(s), tab, max_len);
+}
+
+static void sort_perf(Performance_Entry *pref_entries, int lo, int hi)
+{
+  int i, pivot;
+  
+  if (lo >= hi)
+    return;
+
+  pivot = lo;
+  for (i = lo + 1; i < hi; i++) {
+    if (perf_entries[i].accum < perf_entries[pivot].accum) {
+      Performance_Entry tmp = perf_entries[pivot];
+      perf_entries[pivot] = perf_entries[i];
+      perf_entries[i] = perf_entries[pivot+1];
+      perf_entries[pivot+1] = tmp;
+      pivot++;
+    }
+  }
+
+  sort_perf(perf_entries, lo, pivot);
+  sort_perf(perf_entries, pivot+1, hi);
+}
+
+static void show_perf(Performance_Entry *perf_entries, int perf_count,
+                      int len, int name_len, int gc_len,
+                      int depth)
+{
+  intptr_t total = 0, gc_total = 0;
+  int i, j, k, m, n;
+  char name_tab[16], tab[10], gc_tab[10], pre_indent[8], post_indent[8];
+  Performance_Cat cats[MAX_PERF_CATS];
+  int num_cats = 0;
+
+  memset(cats, 0, sizeof(cats));
+
+  if (!depth) {
+    for (i = 0; i < perf_count; i++) {
+      for (j = 0; recats[j].entry; j++) {
+        if (!strcmp(recats[j].entry, perf_entries[i].name)) {
+          for (m = 0; m < num_cats; m++) {
+            if (!strcmp(recats[j].cat, cats[m].name))
+              break;
+          }
+          if (num_cats <= m) num_cats = m+1;
+          cats[m].name = recats[j].cat;
+          for (k = 0; k < perf_count; k++) {
+            if (perf_entries[k].name) {
+              if (!strcmp(perf_entries[k].name, recats[j].cat))
+                break;
+            } else
+              break;
+          }
+          perf_entries[k].name = recats[j].cat;
+          if (perf_count <= k) perf_count = k+1;
+          perf_entries[k].accum += perf_entries[i].accum;
+          perf_entries[k].gc_accum += perf_entries[i].gc_accum;
+          perf_entries[k].count = -1;
+
+          n = cats[m].perf_count++;
+          cats[m].perf_entries[n] = perf_entries[i];
+          perf_entries[i].accum = 0;
+          perf_entries[i].gc_accum = 0;
+          perf_entries[i].count = 0;
+        }
+      }
+    }
+  }
+
+  sort_perf(perf_entries, 0, perf_count);
+
+  for (i = 0; i < perf_count; i++) {
+    n = strlen(perf_entries[i].name);
+    if (n > name_len) name_len = n;
+    total += perf_entries[i].accum;
+    gc_total += perf_entries[i].gc_accum;
+  }
+
+  n = numlen(total);
+  if (n > len) len = n;
+  n = numlen(gc_total);
+  if (n > gc_len) gc_len = n;
+
+  if (name_len >= sizeof(name_tab))
+    name_len = sizeof(name_tab) - 1;
+  if (len >= sizeof(tab))
+    len = sizeof(tab) - 1;
+  if (gc_len >= sizeof(gc_tab))
+    gc_len = sizeof(gc_tab) -1;
+
+  for (i = 0; i < depth * 2; i++) {
+    pre_indent[i] = ' ';
+  }
+  pre_indent[i] = 0;
+  for (i = 0; i < (3 - depth) * 2; i++) {
+    post_indent[i] = ' ';
+  }
+  post_indent[i] = 0;
+
+  if (!depth)
+    scheme_log(NULL, SCHEME_LOG_ERROR, 0, ";;");
+
+#define BASE_TIMES_TEMPLATE ";; %s%s%s%s  %s%ld [%s%ld] ms"
+#define FULL_TIMES_TEMPLATE BASE_TIMES_TEMPLATE " ; %ld times"
+
+  for (i = 0; i < perf_count; i++) {
+    if (perf_entries[i].count)
+      scheme_log(NULL, SCHEME_LOG_ERROR, 0,
+                 ((perf_entries[i].count < 0)
+                  ? BASE_TIMES_TEMPLATE
+                  : FULL_TIMES_TEMPLATE),
+                 pre_indent,
+                 perf_entries[i].name,
+                 tab_string(perf_entries[i].name, name_tab, name_len),
+                 post_indent,
+                 tab_number(perf_entries[i].accum, tab, len),
+                 perf_entries[i].accum,
+                 tab_number(perf_entries[i].gc_accum, gc_tab, gc_len),
+                 perf_entries[i].gc_accum,
+                 perf_entries[i].count);
+    for (m = 0; m < num_cats; m++) {
+      if (!strcmp(perf_entries[i].name, cats[m].name))
+        show_perf(cats[m].perf_entries, cats[m].perf_count, len, name_len, gc_len, depth+1);
+    }
+  }
+
+  if (!depth) {
+    scheme_log(NULL, SCHEME_LOG_ERROR, 0,
+               ";; %stotal%s  %s%ld [%s%ld] ms",
+               tab_number(total, tab, len),
+               tab_string("total", name_tab, name_len),
+               post_indent,
+               total,
+               tab_number(gc_total, gc_tab, gc_len),
+               gc_total);
+#ifdef MZ_PRECISE_GC
+    scheme_log(NULL, SCHEME_LOG_ERROR, 0, ";;");
+    scheme_log(NULL, SCHEME_LOG_ERROR, 0,
+               ";; [JIT code: %d procs  %d bytes  code+admin: %d bytes]",
+               scheme_code_count,
+               scheme_code_total,
+               scheme_code_page_total);
+#endif
+  }
+}
+
+static void show_all_perf()
+{
+  return show_perf(perf_entries, perf_count, 0, 0, 0, 0);
+}
+
+void scheme_performance_record_start(GC_CAN_IGNORE Scheme_Performance_State *perf_state)
+{
+#if defined(MZ_USE_PLACES)
+  if (scheme_current_place_id != 0)
+    return;
+#endif
+  
+  if (!perf_reg) {
+    if (scheme_getenv("PLT_LINKLET_TIMES")) {
+      perf_reg = 1;
+      scheme_atexit(show_all_perf);
+    } else {
+      perf_reg = -1;
+    }
+  }
+
+  if (perf_reg < 0)
+    return;
+
+  perf_state->gc_start = scheme_total_gc_time;
+  perf_state->start = scheme_get_process_milliseconds();
+  perf_state->old_nested_delta = nested_delta;
+  perf_state->old_nested_gc_delta = nested_gc_delta;
+
+  nested_delta = 0;
+  nested_gc_delta = 0;
+}
+
+void scheme_performance_record_end(const char *who, GC_CAN_IGNORE Scheme_Performance_State *perf_state)
+{
+  int i;
+  intptr_t d, gc_d;
+  Scheme_Performance_State zero_perf_state;
+
+#if defined(MZ_USE_PLACES)
+  if (scheme_current_place_id != 0)
+    return;
+#endif
+
+  if (perf_reg < 0)
+    return;
+
+  for (i = 0; i < MAX_PERF_ENTRIES; i++) {
+    if (perf_entries[i].name) {
+      if (!strcmp(perf_entries[i].name, who))
+        break;
+    } else
+      break;
+  }
+
+  if (i >= MAX_PERF_ENTRIES)
+    return;
+
+  if (!perf_state) {
+    memset(&zero_perf_state, 0, sizeof(zero_perf_state));
+    perf_state = &zero_perf_state;
+  }
+
+  d = (scheme_get_process_milliseconds() - perf_state->start);
+  gc_d = (scheme_total_gc_time - perf_state->gc_start);
+
+  perf_state->old_nested_delta += d;
+  perf_state->old_nested_gc_delta += gc_d;
+
+  d -= nested_delta;
+  gc_d -= nested_gc_delta;
+
+  nested_delta = perf_state->old_nested_delta;
+  nested_gc_delta = perf_state->old_nested_gc_delta;
+
+  if (!perf_entries[i].name) {
+    perf_entries[i].name = who;
+    perf_count++;
+  }
+  perf_entries[i].accum += d;
+  perf_entries[i].gc_accum += gc_d;
+  perf_entries[i].count++;
+}
+
 /*========================================================================*/
 /*                         precise GC traversers                          */
 /*========================================================================*/
